#! /usr/local/bin/perl -w

# 2024, Tue 6 Jun 00 (PST), CAIDA
#
# Nevil Brownlee, ITSS Technology Development, The University of Auckland

# Makes gnuplot files for a distribution from a .dif or .sum file
#
# Command line: 
#   ./plt_dist.pl  dns-root.dif xxx.plt  flowkind=7 @time ^ % toturnaroundtime1
#          from_file to_file  attrib_spec  @which_stats ^! +%/ distribs required
#
#   attrib spec selects values, e.g. flowtype=1..3 sourcetranstype=17
#   valid operators are ==, != (specifying tests); only one range is allowed
#   may also specify a list of values, e.g. sourcetranstype = 1,6,17
#
#   @time in the specifiers plots distrib for that time, e.g. "0617.1800"
#
#   ^ specifier for 'multi-range,' i.e. distribs in the list have
#       different ranges, have to make 'values' col for each distrib
#   !  build cumulative % distributions
#   ?  build complement of cumulative % distributions
#         (? means something to sh, have to put it in ''s
#
#   +  Raw distributions (just the counts)
#   %  Percentage distributions (0..100%)
#   /  % distributions, but ignore the overflow bucket
#
# CAUTION: Col 1 of the file is $x, the sample time
#          Col 2 is the distribution's x value FOR ALL FOLLOWING COLUMNS
#          => all distribs must use same scale (bins, transform, scale, lo, high)


use strict;
use Time::Local;

local $::ln10 = 2.3025850929940457;

local %::month = (
      "Jan" =>  1, "Feb" =>  2, "Mar" =>  3, "Apr" =>  4,
      "May" =>  5, "Jun" =>  6, "Jul" =>  7, "Aug" =>  8,
      "Sep" =>  9, "Oct" => 10, "Nov" => 11, "Dec" => 12,
   );

my ($c_interval);  # Collection interval (from fdf ## record)

my ($from_file, $to_file);

$from_file = shift;  # First item from command line
if (defined($from_file)) {
   open(FROM, $from_file)
   or die "Can't open from_file: $!\n";
   }
else { printf "No from_file specified!\n"; }
 
$to_file = shift;  # Second item from command line
if (defined($to_file)) {
   open(TO, ">>$to_file")
   or die "Can't open to_file: $!\n";
   }
else { printf "No to_file specified!\n"; }

my $ntm_hdr = <FROM>;  # Read the input header record
if ($ntm_hdr =~ /-c(\d+) /) {
   $c_interval = $1;
   }
print TO $ntm_hdr;

my $format;
for (;;) {  #  Find the #Format record
   $format = <FROM>;
   print TO $format;
   last if $format =~ /^#Format: /
   }
$format =~ s/^#Format: //;

my ($i, %format, %dist, $next);
my ($nf, @type, @dif);
# type: 0 = scalar, 1 = tcpdata, 2 = distribution
# dif:  1 = dif (to be summed), 0 = fixed

my @atname;
for ($nf = 0; $format =~ s/([\w\[\(]\w*[\]|\)]?)//; $nf++) {  # Split up fmt
   $next = $1;
   $type[$nf] = 0;
   if ($next =~ s/^\((.+)\)$/$1/ ) {
      $dist{$next} = $nf;
      $type[$nf] = 2;  # Distribution
      }
   elsif ($next =~ s/^\[(.+)\]$/$1/ ) {
      $type[$nf] = 1;  # TCPdata
      }
   $format{$next} = $nf;
   $atname[$nf] = $next;  ###
   $dif[$nf] = 0;
   if ($next =~ /^d_/) {
      $dif[$nf] = 1;  # To be summed
      }
   }
for (my $j = 0; $j != $nf; ++$j) {
    print "$j  $atname[$j]  $type[$j]  $dif[$j]\n";
   }


my ($c1,$c2);  # c1..c2 columns written to stats file
my $ptime;  # Plot the distrib(s) for this time

my ($item, @attrb, @attst, @attvl, @attvh, @range);
my %values;  # Hash to map alist of values
my $multi_range = 0;  # 1 => produce 'values' column for each distrib
my $cumulative = 0;   # 1 => produce cumulative distributions(s)
my $complementary = 0;   # 1 => produce complementary cumulative distributions(s)

my $errs = 0;

for (my $a = 0; $item = shift; ) {  # Grab items from command line

   last if $item eq '+';  # Stop when we find a '+'
   last if $item eq '%';  #    or a '%'
   last if $item eq '/';  #    or a '/'
   $item = lc($item);

   if ($item eq '!') {
      $cumulative = 1;
      next;
      }
   if ($item eq '?') {
      $cumulative = $complementary = 1;
      next;
      }

   if ($item eq '^') {
      $multi_range = 1;
      next;
      }

   if ($item =~ s/^@//) {  # @ introduces a plot time
      $ptime = $item;
      print " ptime = $ptime\n";
      next;
      }

   my ($atr, $test, $val, $valh, $range);
   if ($item =~ s/\.\.(\d+)//) { 
      $valh = $1;  # Range
      $range = '..';
      }
   elsif ($item =~ s/\,(.+)$//) {
      $valh = $1;  # List
      $range = ',';
      }
   else { 
      $valh = '';  $range = '';
      }

   if ($item =~ /^(\w+)([=!\#]+)([\w.]+)/) {
      $atr = $1;  $test = substr($2,0,1);  $val = $3;
      if ($range eq '..') {
         $c1 = $val;  $c2 = $valh;
         }
      elsif ($range eq ',') {
         $values{$val} = $c1 = $c2 = 1;
         for (my $vh = $valh . ',';  $vh =~ s/^(\d+)\,//; ) {
            $values{$1} = ++$c2;
            }
         }
      else {  # 0 or 1 value specified
         $c1 = $c2 = $val;
         }
      }
   else { $atr = $item;  $test = '=';  $val = ''; }

   if (defined $format{$atr} ) {
      $attrb[$a] = $atr;  $attst[$a] = $test;
      $attvl[$a] = $val;  $attvh[$a] = $valh;  $range[$a] = $range;
      printf " %s %s %s %s %s\n", $atr, $test, $val, $range,
         $valh ne '' ? $valh : '';
      ++$a;
      }
   else {
      print STDERR "unknown attrib '$atr'\n";
      ++$errs;
      }
   }

my @plot;

if (defined($item)) {
   for (my $p=0; $plot[$p] = shift; ++$p ) { # Grab distribs
      $plot[$p] = lc($plot[$p]);
      if (defined $format{$plot[$p]}) { 
         printf " %s", $plot[$p];
         }
      else { 
         print STDERR "Unknown distribution '$plot[$p]' <<<\n";
         ++$errs;
         }
      }
   }
pop @plot;   # Last shift will have returned NULL
print "\n";

print TO "##plt_dist.pl  $from_file ";
for (my $j = 0; $j != scalar(@attrb); ++$j) {
   print TO " $j: $attrb[$j]$attst[$j]$attvl[$j]";
   print " key $j: $attrb[$j]$attst[$j]$attvl[$j]";
   if ($attvh[$j] ne '') {
      print TO "$range[$j]$attvh[$j]";
      print  "$range[$j]$attvh[$j]";
      }
   print "\n";
   }

print TO ' ';
for (my $p=0; defined($plot[$p]); ++$p ) {
   printf TO " %s", $plot[$p];
   }
print TO "\n";

if ($errs != 0) { exit; }

if (defined($c1) && defined($c2) && $c2 >= $c1) {
   my $n_distribs = scalar(@plot);
   if ($n_distribs != 1) {
      $c2 = $c1-1 + ($c2-$c1+1)*$n_distribs;
      }
   print "$n_distribs distributions => plot columns $c1 to $c2\n";
   }
else  {
   print "Attribute range for plot file columns not properly specified !!!\n";
   exit;
   }


# %format       = index in records for attribtes (from #Format)
# @attrb,@attvl = array of attrib/value pairs for selecting flows
# $item         = +  Use distrib raw counts (i.e. 'as is')
#                 %  Convert each sample to pecentages
#                 /  Convert to percentages, but ignore oflo bucket
# @plot         = list of attribs to plot

#printf "attrib %s=%s, plot %s=%d\n",
#    $attrb[0], $attvl[0], $plot[0], $format{$plot[0]};

my $nbr = 0;

#                Globals for get_time()
my $c_time;      # Collection time (seconds from 1970)
my $last_c_time = 0;
my $delta = 0;   # Actual interval - sample interval (0 for regular samples)
my $x;           # x value (time of day) for gnuplot

my %flows = ();  # Hash for flows in sample

my @distrib;  # Distributions for current sample
my @tolost;   # Lost request packet counts for current sample

my $tolostx = $format{"d_tolostpdus"};
my $pltx = $format{$plot[0]};  # Index in record of attrib to plot


my $gap_to = 0;

while (<FROM>) { # read  flows file

   chop;    # trim new line
   if (/^#/) { # starts with a #
      my $hrec = $_;
      if (s/^#Time: //) {
         get_time();
         if ($delta < -5 || $delta > 5) {  # Meter restart || missing collection
            if (!$gap_to) {
               print TO "\n";  # Blank line causes gap in gnuplot 'lines'
               $gap_to = 1;
               }
            }
	 } 
      elsif (s/^#EndData: //) {
         if ($x eq $ptime) {
#$print "*** x=$x eq ptime=$ptime\n";
            my $n_flows = 0;
            my $col;
            foreach my $key (sort keys %flows) {
               my @f = @{ $flows{$key} };
               my $sv = $f[$format{$attrb[0]}];  # Value of selector attribute
               if ($range[0] eq ',')
                  { $sv = $values{$sv}; }
               $col = $c1 + ($sv-$c1) * scalar(@plot);
#print "av=$f[$format{$attrb[0]}], sv=$sv, c1=$c1, c2=$c2, col=$col\n";
               foreach my $plt (@plot) {
                my @vr = $f[$format{$plt}];
#$print "??? plt=$plt, Upper=$vr[5] $vr[6]\n";
                  $distrib[$col] = $f[$format{$plt}];
                  if (defined($tolostx)) {
                     $tolost[$col] = $f[$tolostx];
                     }
                  ++$col;
                  } 
               ++$n_flows;
	       }
            if ($n_flows == 0) {
               if (!$gap_to) {
                  print TO "\n";  # Blank line causes gap in gnuplot 'lines'
                  $gap_to = 1;
                  }
               }
            else {
               put_distribs(\*TO, $x, $c1,$c2, $item);
               exit;
               $gap_to = 0;
               }
            @tolost = @distrib = ();  # Clear last sample's data
            %flows = ();
	   }
        }

      next;
      }  # End of # record handling

   my $vref = get_flow();

   my $key = match_flow($vref);  # '' if we don't want the flow
   if ($key ne '') {
#print "x=$x, key=$key\n";
      $flows{$key} = $vref;
      }

#   ++$nbr;
#   if ($nbr > 50) { exit; }
   }

# That's it
#--------------------------------------------------------------------------


sub match_flow{  # Returns flow key, '' if match fails
   my ($vref) = @_;

   my ($av, $match, $fail);
   $fail = 0;
   for (my $q = 0; $q != scalar(@attrb); ++$q) {
      if ($attvl[$q] ne '') {
         next if $attst[$q] eq '#';
         $av = $vref->[$format{$attrb[$q]}];
         if ($range[$q] eq ',')
            { $match = defined($values{$av}); }
         elsif ($attvh[$q] eq '')
            { $match = $av == $attvl[$q]; }
         else {
            $match = ($av >= $attvl[$q] && $av <= $attvh[$q]);
            }
         if ($match) {
            last if $fail = $attst[$q] eq '!';
            }
         else {
            last if $fail = $attst[$q] eq '=';
            }
         } 
      }

   $av = $vref->[$format{$attrb[0]}];
   if ($range[0] eq ',')
      { $av = $values{$av}; }
   return $fail ? '' :
      sprintf("$attrb[0] %03d", $av);
   }

sub get_flow{  # Take $_, return ref to array of values

   my @value = ();
   for ($i = 0;  s/([0123456789ABCDEF.-]+|\(|\[)//;  $i++) {
         # remove leading non-numerics followed by (, [ or a value
         # NOTE: .- appear in addresses, : appears in V6 addresses
      if ($1 eq '(') {  #  have distrib
	 s/\W*([^)]+)\)//;  # remove everything upto next ) save it in $1
	 my @v = split(/ +/, $1);   # put the distribution values in @v
	 my $v = {};                # $v points to a hash
	 if (($v->{Trans} = $v[0]) != 0 ) {  # construct structure
	    $v->{Scale} = $v[1];
	    $v->{Lower} = $v[2];
	    $v->{Upper} = $v[3];
	    $v->{Buckets} = $v[4];
	    $v->{P1} = $v[5];
	    $v->{P2} = $v[6];
	    $v->{P3} = $v[7];
	    $v->{Dist} = [splice(@v,8)]; # copies from element 8 to end
#            if ($v->{P1} != 0) {
#               $samples_per_int = $c_interval/$v->{P1};
#               $v->{Dist}->[0] %= $samples_per_int;
#                  # Ignore empty collection intervals
#               }
            $value[$i] = $v;
	    }
	 } 
      elsif ($1 eq '[') {  # TCPdata
	 s/([^\]]+)\]//;  # Remove everything upto next ], save it in $1 
         my @v = split(/ +/, $1);
	 $value[$i] = \@v;
         }
      else {
	 $value[$i] = $1;
	 }
      }
   return \@value;  # Return ref to array
   }
 

sub put_flow{  # Write output record
   my ($FILE, $vref) = @_;

   for (my $j = 0; $j != $nf; ++$j) {
      if ($type[$j] == 0) {  # Scalar
         print $FILE "$vref->[$j] ";
         }
      elsif ($type[$j] == 2) {  # Distribution
         my $hr = $vref->[$j];
         if (!defined($hr->{Trans})) {
            print $FILE "(0) ";  # Empty distribution
	    }
	 else {
            print $FILE 
               "($hr->{Trans} $hr->{Scale} $hr->{Lower} $hr->{Upper}  ";
            print $FILE 
               "$hr->{Buckets} $hr->{P1} $hr->{P2} $hr->{P3}  ";
            my @bucket = @{ $hr->{Dist} };
            print $FILE "@bucket) ";
            }
         }
      else {  # TCPdata
         my @data = @{ $vref->[$j] };
         print $FILE "[@data] ";
         }
      }
   print $FILE "\n";
   }


sub get_time{  # Take #Time from $_, return $x, $delta, $last_ctime
   if (/(^\d+):(\d+):(\d+) +\w+ +(\d+) +(\w+) +(\d+) +/) {
      #    HH    MM   SS    day    dd    mm     yyyy
#      $c_time = timegm($3,$2,$1, $4,$::month{$5}-1,$6-1900);
      $c_time = timelocal($3,$2,$1, $4,$::month{$5}-1,$6-1900);
#      print $_, "  ", $c_time, " ", $last_c_time, "\n";
      if ($last_c_time != 0) {
         my $interval = $c_time - $last_c_time;
         $delta = $interval - $c_interval;
         if ($delta < -2 || $delta > 2) {
            print STDERR $_, "\n";
            print STDERR "   interval = $interval <<<<<<<<<\n";
            }
         }
      $last_c_time = $c_time;
      $x = sprintf("%02d%02d.%02d%02d", $::month{$5},$4, $1,$2);
      }               # mmdd.HHMM
#      $x = sprintf("%04d%02d%02d:%02d%02d", $6,$::month{$5},$4, $1,$2);
#      }               # YYYYmmdd:HHMM
   }

sub put_distribs {
   my($FILE, $x, $first,$last, $percent) = @_;  # Arguments

   my ($ld, $dist, $count);
   my (@div, @low, @nb, @step, @pcmult, @total, @sum);

   for ($ld = $first; $ld <= $last; ) {
      foreach my $plt (@plot) {
#print "plt=$plt, fmt=$format{$plt}\n";
         if ($plt =~ /pdus|packetsize|ime4/) {
           $div[$ld] = 1;  # pkts, bytes, train length
            }
         elsif ($plt =~ /bitrate/) {
            $div[$ld] =  1000000;  # Mbps (rather than kbps)
            }
         else {
            $div[$ld] = 1000;  # kB/s, ms, kB ..
            }
         ++$ld;
         }
      }

   for ($count = 0, $ld = $first; $ld <= $last; ++$ld) {
      if (defined $distrib[$ld]) {
         my ($scale, $low, $high, $nb, $step);
         $dist = $distrib[$ld];
         my $ttt = $dist->{Trans};
#$print "--- ld=$ld, count=$count, trans=$ttt\n";
         if ($dist->{Trans} == 0) { next; }
         ++$count;
         $scale =  $dist->{Scale};
         $low = $dist->{Lower}*10**$scale;
         $high = $dist->{Upper}*10**$scale;
         $nb = $dist->{Buckets};
         if ($percent eq '/') {  # Ignore the 'overflow' counter
	    --$nb;
	    }

         if ($dist->{Trans} == 1) {
            $step = ($high - $low) / ($nb - 1);
            }
         else {
            $step = (log($high) - log($low)) / (($nb - 1)*$::ln10);
            }

         if ($ld != $first) {
            if ($nb != $nb[$first]) {
               print "distrib[$ld] has different buckets to distrib$first] <<<\n";
               }
            if (!$multi_range && ($low != $low[$first] || $step != $step[$first])) {
               print "different ranges!  Use multi_range (^ on command line) <<<\n";
               }
            }
         $low[$ld] = $low;  $nb[$ld] = $nb;  $step[$ld] = $step;
#print "multi=$multi_range, ld=$ld, low=$low, high=$high, nb=$nb,step=$step\n";
         }
      }

   if ($count == 0) {
      printf $FILE "#$x: no distribs in sample\n";
      return;
      }

   if ($percent eq '%' || $percent eq '/' || $cumulative || $complementary) {
      for (my $j = $first; $j <= $last; ++$j) {
	 if (defined $distrib[$j]) {
            $dist = $distrib[$j];
            my $tot = 0;  # Compute normalising factor
            for (my $k = 0; $k <= $nb[$j]; $k++) {  
               $tot += $dist->{Dist}->[$k];
               }
            my $plot_max = $complementary ? 1 : 100;
            $pcmult[$j] = $tot != 0 ? $plot_max/$tot : 1;
            $total[$j] = $tot;
	    if ($tot == 0) {
#$print ">>> col=$j, tot=0\n";
               $total[$j] = 1;  # Only used for cummulative
                                # Avoid divide by zero
	       }
	    }
         else {
            $pcmult[$j] = 1;
            }
         }
      }
   else {
      for (my $j = $first; $j <= $last; ++$j) {
         $pcmult[$j] = 1;
         }
      }

#printf "scale=%d, low=%f, high=%f, div=%d, nb=%d\npcmult=",
#   $scale, $low, $high, $div, $nb;
#for (my $j = $first; $j <= $last; ++$j) {
#   if (defined $distrib[$j]) {
#      printf " %d=%.3f, ", $j,$pcmult[$j];
#      }
#   else {
#       printf " %d=-1, ", $j;
#      }
#   }
#printf "\n";

   for (my $row = 0; $row <= $nb[$first]; ++$row) {
      printf $FILE "%s", $x;

      for (my $j = $first; $j <= $last; ++$j) {
         if ($multi_range || $j == $first) {
            my $d = ($dist->{Trans} == 1 ? $low[$j] + $row*$step[$j] :
                  $low[$j] * 10**($step[$j]*$row)) / $div[$j];
            printf $FILE " %.2f", $d;
            }
	 if (defined $distrib[$j]) {
            if ($cumulative) {
               $sum[$j] += $distrib[$j]->{Dist}->[$row];
               my $plot_max = $complementary ? 1 : 100;
               my $prob = $sum[$j]*$plot_max/$total[$j];
               if ($complementary) { $prob = 1 - $prob; }
               printf $FILE " %f", $prob;
#               printf $FILE " %.2f", $prob;
               }
            else {
               printf $FILE " %.2f", $distrib[$j]->{Dist}->[$row] * $pcmult[$j];
               }
	    }
         else {
	     printf $FILE " ?";
            }
         }

      printf $FILE "\n";
      }
   print $FILE "\n";  # Blank line to separate distribs
  }
