#! /usr/local/bin/perl -w

# 1335, Tue 5 Dec (PST), CAIDA
#
# Nevil Brownlee, The University of Auckland / CAIDA

# Makes gnuplot files for a distribution from a .dif or .sum file
#
# Command line: 
#   ./plt_stat.pl  dns-root.dif xxx.plt  flowtype=7 @stat-rq % toturnaroundtime1
#          from_file to_file  attrib_spec  @which_stats  +%/ distribs required
#
#   attrib spec selects values, e.g. flowtype=1..3 sourcetranstype=17
#   valid operators are ==, != (specifying tests); only one range is allowed

use strict;
use Time::Local;

local $::ln10 = 2.3025850929940457;

local %::month = (
      "Jan" =>  1, "Feb" =>  2, "Mar" =>  3, "Apr" =>  4,
      "May" =>  5, "Jun" =>  6, "Jul" =>  7, "Aug" =>  8,
      "Sep" =>  9, "Oct" => 10, "Nov" => 11, "Dec" => 12,
   );

my ($c_interval);  # Collection interval (from fdf ## record)

my ($from_file, $to_file);

$from_file = shift;  # First item from command line
if (defined($from_file)) {
   open(FROM, $from_file)
   or die "Can't open from_file: $!\n";
   }
else { printf "No from_file specified!\n"; }
 
$to_file = shift;  # Second item from command line
if (defined($to_file)) {
   open(TO, ">>$to_file")
   or die "Can't open to_file: $!\n";
   }
else { printf "No to_file specified!\n"; }

my $ntm_hdr = <FROM>;  # Read the input header record
if (!defined($ntm_hdr)) {
   print "read from $from_file falied!\n";
   exit;
   }
if ($ntm_hdr =~ /-c(\d+) /) {
   $c_interval = $1;
   }
print TO $ntm_hdr;

my $format;
for (;;) {  #  Find the #Format record
   $format = <FROM>;
   print TO $format;
   last if $format =~ /^#Format: /
   }
$format =~ s/^#Format: //;

my ($nf, %format, %dist, @type, @dif);
# type: 0 = scalar, 1 = tcpdata, 2 = distribution
# dif:  1 = dif (to be summed), 0 = fixed

unpack_format_record($format);

#  Statistics required (in the order put_stats() will write them to .plt filke)
#     Specify which you want with binary stringin <>.  Default is <1100000>

my $mdrq = 1;  #  1 Median
my $lprq = 1;  #  2 Loss percentage
my $lorq = 0;  #  3 First quartile
my $hirq = 0;  #  4 Third quartile
my $mnrq = 0;  #  5 Min
my $avrq = 0;  #  6 Mean
my $mxrq = 0;  #  7 Max
my $p5rq = 0;  #  8 5%-ile
my $p9rq = 0;  #  9 95%-ile
my $tcrq = 0;  # 10 Total count
my $tlrq = 0;  # 11 ToLostPDUs
my $scrq = 0;  # 12 Scalar(s) reqd.  CCAUTION: can't also have distribs!

my $lo_p = 25;  my $hi_p = 75;  my $min_pts = 8;  # Hi/lo percentiles
my $min_5pts = 40;

my ($c1,$c2);  # c1..c2 columns written to stats file

my ($item, @attrb, @attst, @attvl, @att_str, @attvh, @nvalues, @amult);
my $errs = 0;

for (my $a = 0; $item = shift; ) {  # Grab items from command line

   last if $item eq '+';  # Stop when we find a '+'
   last if $item eq '%';  #    or a '%'
   last if $item eq '/';  #    or a '/'
   $item = lc($item);

   if ($item =~ /^@/) {  # Stats-reqd specified as a binary string after @
      my ($d, $len);
      $len = length($item);
      if ($len >= 2)
	  { $mdrq = substr($item,1,1); }
      if ($len >= 3)
	  { $lprq = substr($item,2,1); }
      if ($len >= 4)
	  { $lorq = substr($item,3,1); }
      if ($len >= 5)
	  { $hirq = substr($item,4,1); }
      if ($len >= 6)
	  { $mnrq = substr($item,5,1); }
      if ($len >= 7)
	  { $avrq = substr($item,6,1); }
      if ($len >= 8)
	  { $mxrq = substr($item,7,1); }
      if ($len >= 9)
	  { $p5rq = substr($item,8,1); }
      if ($len >= 10)
	  { $p9rq = substr($item,9,1); }
      if ($len >= 11)
	  { $tcrq = substr($item,10,1); }
      if ($len >= 12)
	  { $tlrq = substr($item,11,1); }
      if ($len >= 13)
	  { $scrq = substr($item,12,1); }
      next;
      }

   my ($atr, $test, $val, $str_atr, $valh);

   if ($item =~ s/\.\.(\d+)//) { 
      $valh = $1;
      }
   else { $valh = ''; }
   if ($item =~ /^(\w+)([=!\#]+)([\w.]+)/) {
      $atr = $1;  $test = substr($2,0,1);  $val = $3;
      $str_atr = $val =~ /\./ ;
      if ($valh ne '' && $val > $valh) {
         print "Bad attribute range for $atr ($val > $valh)\n";
         ++$errs;
         }
      }
   else { $atr = $item;  $test = '=';  $val = ''; }

   if (defined $format{$atr} ) {
      $attrb[$a] = $atr;  $attst[$a] = $test;
      $attvl[$a] = $val;  $att_str[$a] = $str_atr;
      $attvh[$a] = $valh;
      $nvalues[$a] = $valh ne '' ? $valh-$val+1 : ($val ne '' ? 1 : 0);
      printf " %s %s %s %s (%d values)\n", $atr, $test, $val,
         $valh ne '' ? ".. $valh" : '', $nvalues[$a];
      ++$a;
      }
   else {
      print STDERR "Unknown attribute '$atr'\n";
      ++$errs;
      }
   }

my $nstats = 0;
if ($mdrq) { ++$nstats; }
if ($lprq) { ++$nstats; }
if ($lorq) { ++$nstats; }
if ($hirq) { ++$nstats; }
if ($mnrq) { ++$nstats; }
if ($avrq) { ++$nstats; }
if ($mxrq) { ++$nstats; }
if ($p5rq) { ++$nstats; }
if ($p9rq) { ++$nstats; }
if ($tcrq) { ++$nstats; }
if ($tlrq) { ++$nstats; }
if ($scrq) { ++$nstats; }

my @plot;

if (defined($item)) {  # Stopped on +%/ (rather than \n)
   for (my $p=0; $plot[$p] = shift; ++$p ) { # Grab distribs
      $plot[$p] = lc($plot[$p]);
      if (defined $format{$plot[$p]}) { 
         printf " %s", $plot[$p];
         }
      else { 
         print STDERR "Unknown distribution '$plot[$p]' <<<\n";
         ++$errs;
         }
      }
   }
pop @plot;   # Last shift will have returned NULL
print "\n";

if ($errs != 0) { exit; }

my $n_distribs = scalar(@plot);
my $n_cols = $n_distribs;
for (my  $j = scalar(@attrb)-1; $j >= 0; --$j) {
   $amult[$j] = $n_cols;
   $n_cols *= $nvalues[$j];
   }

print TO "##plt_stat  $from_file ";
for (my $j = 0; $j != scalar(@attrb); ++$j) {
   print TO " $attrb[$j]$attst[$j]$attvl[$j]";
   print " key $j: $attrb[$j]$attst[$j]$attvl[$j]";
   if ($attvh[$j] ne '') {
      print TO "..$attvh[$j]";
      print  "..$attvh[$j]";
      print " [*$amult[$j]]";
      }
   print "\n";
   }
print "$n_cols columns per row\n";
print TO "  \@$mdrq$lprq$lorq$hirq$mnrq$avrq$mxrq$p5rq$p9rq$tcrq$tlrq  " .
   "stats=$nstats distribs=$n_distribs cols=$n_cols\n";
print TO "\n";

# %format       = index in records for attribtes (from #Format)
# @attrb,@attvl = array of attrib/value pairs for selecting flows
# $item         = +  Use distrib raw counts (i.e. 'as is')
#                 %  Convert each sample to pecentages
#                 /  Convert to percentages, but ignore oflo bucket
# @plot         = list of attribs to plot

my $nbr = 0;  # For counting records when debugging

#                Globals for get_time()
my $c_time;      # Collection time (seconds from 1970)
my $last_c_time = 0;
my $delta = 0;   # Actual interval - sample interval (0 for regular samples)
my $x;           # x value (time of day) for gnuplot

my %flows = ();  # Hash for flows in sample

my @distrib;  # Distributions for current sample
my @tolost;   # Lost request packet counts for current sample

my $tolostx = $format{"d_tolostpdus"};

my $gap_to = 0;

while (<FROM>) { # read  flows file

   chop;    # trim new line
   if (/^#/) { # starts with a #
      my $hrec = $_;
      if (s/^#Time: //) {
         get_time();
         if (abs($delta) > 5) {  # Meter restart || missing collection
            if (!$gap_to) {
               print TO "\n";  # Blank line causes gap in gnuplot 'lines'
               $gap_to = 1;
               }
            }
	 } 
      elsif (s/^#EndData: //) {
         my $n_flows = 0;
         my $col;

         foreach my $key (sort keys %flows) {
            my @f = @{ $flows{$key} };  # Actual flow
            my @kv = split(/ /, $key);
            $col = 0;
            for (my $q = 0; $q != scalar(@attrb); ++$q) {
               my $sv = $kv[$q];  # Value of selector attribute
               $col += ($sv-$attvl[$q])*$amult[$q];
               }
            foreach my $plt (@plot) {
               $distrib[$col] = $f[$format{$plt}];
               if (defined($tolostx)) {
                  $tolost[$col] = $f[$tolostx];
                  }
               ++$col;
               } 
            ++$n_flows;
	    }
         if ($n_flows == 0) {
            if (!$gap_to) {
               print TO "\n";  # Blank line causes gap in gnuplot 'lines'
               $gap_to = 1;
               }
            }
         else {
            if ($scrq) {
               put_scalars(\*TO, $x, 0,$n_cols-1);
               }
            else {
               put_stats(\*TO, $plot[0], $x, 0,$n_cols-1, $item);
               $gap_to = 0;
               }
            }
         @tolost = @distrib = ();  # Clear last sample's data
         %flows = ();
         }
      elsif (s/^#Format: //) {
         $hrec =~ s/^#Format: //;
         unpack_format_record($hrec);
         $tolostx = $format{"d_tolostpdus"};
         for (my $p = 0; $p != scalar(@plot); ++$p ) { # Grab distribs
            if (!defined $format{$plot[$p]}) { 
               print STDERR
        "Format change, last time $x: nknown distribution '$plot[$p]' <<<\n";
               ++$errs;
               }
            }
         }

      next;
      }  # End of # record handling

   my $vref = get_flow();
   my $key = match_flow($vref);  # '' if we don't want the flow
   if ($key ne '') {
      $flows{$key} = $vref;
      }

#   ++$nbr;
#   if ($nbr > 15) { exit; }
   }

print "Last collection time in file was:  $x\n";

# That's it
#--------------------------------------------------------------------------

sub unpack_format_record {
   my ($fmt) = @_;

   $nf = 0;  %format = ();  %dist = (); @type = ();  @dif  = ();

   for ($nf = 0; $fmt =~ s/([\w\[\(]\w*[\]|\)]?)//; ++$nf) {  # Split up fmt
      my $next = $1;
      $type[$nf] = 0;
      if ($next =~ s/^\((.+)\)$/$1/ ) {
         $dist{$next} = $nf;
         $type[$nf] = 2;  # Distribution
         }
      elsif ($next =~ s/^\[(.+)\]$/$1/ ) {
         $type[$nf] = 1;  # TCPdata
         }
      $format{$next} = $nf;
      $dif[$nf] = 0;
      if ($next =~ /^d_/) {
         $dif[$nf] = 1;  # To be summed
         }
      }
   }

sub match_flow{  # Returns flow key, '' if match fails
   my ($vref) = @_;

   my ($av, $match, $fail, $key);
   $fail = 0;  $key = '';
   for (my $q = 0; $q != scalar(@attrb); ++$q) {
      $av = $vref->[$format{$attrb[$q]}];
      $key = $key . ($att_str[$q] ? $av : sprintf("%03d ", $av));
      if ($attvl[$q] ne '') {
         if ($attvh[$q] eq '') {
            $match = $att_str[$q] ? $av eq $attvl[$q] : $av == $attvl[$q];
            }
         else {
            $match = ($av >= $attvl[$q] && $av <= $attvh[$q]);
            }
         if ($match) {
            last if $fail = $attst[$q] eq '!';
            }
         else {
            last if $fail = $attst[$q] eq '=';
            }
         } 
      }
   return $fail ? '' : $key;
   }

sub get_flow{  # Take $_, return ref to array of values

   my @value = ();
   for (my $i = 0;  s/([0123456789ABCDEF.-]+|\(|\[)//;  $i++) {
         # remove leading non-numerics followed by (, [ or a value
         # NOTE: .- appear in addresses, : appears in V6 addresses
      if ($1 eq '(') {  #  have distrib
	 s/\W*([^)]+)\)//;  # remove everything upto next ) save it in $1
	 my @v = split(/ +/, $1);   # put the distribution values in @v
	 my $v = {};                # $v points to a hash
	 if (($v->{Trans} = $v[0]) != 0 ) {  # construct structure
	    $v->{Scale} = $v[1];
	    $v->{Lower} = $v[2];
	    $v->{Upper} = $v[3];
	    $v->{Buckets} = $v[4];
	    $v->{P1} = $v[5];
	    $v->{P2} = $v[6];
	    $v->{P3} = $v[7];
	    $v->{Dist} = [splice(@v,8)]; # copies from element 8 to end
#            if ($v->{P1} != 0) {
#               $samples_per_int = $c_interval/$v->{P1};
#               $v->{Dist}->[0] %= $samples_per_int;
#                  # Ignore empty collection intervals
#               }
            $value[$i] = $v;
	    }
	 } 
      elsif ($1 eq '[') {  # TCPdata
	 s/([^\]]+)\]//;  # Remove everything upto next ], save it in $1 
         my @v = split(/ +/, $1);
	 $value[$i] = \@v;
         }
      else {
	 $value[$i] = $1;
	 }
      }
   return \@value;  # Return ref to array
   }
 

sub put_flow{  # Write output record
   my ($FILE, $vref) = @_;

   for (my $j = 0; $j != $nf; ++$j) {
      if ($type[$j] == 0) {  # Scalar
         print $FILE "$vref->[$j] ";
         }
      elsif ($type[$j] == 2) {  # Distribution
         my $hr = $vref->[$j];
         if (!defined($hr->{Trans})) {
            print $FILE "(0) ";  # Empty distribution
	    }
	 else {
            print $FILE 
               "($hr->{Trans} $hr->{Scale} $hr->{Lower} $hr->{Upper}  ";
            print $FILE 
               "$hr->{Buckets} $hr->{P1} $hr->{P2} $hr->{P3}  ";
            my @bucket = @{ $hr->{Dist} };
            print $FILE "@bucket) ";
            }
         }
      else {  # TCPdata
         my @data = @{ $vref->[$j] };
         print $FILE "[@data] ";
         }
      }
   print $FILE "\n";
   }


sub get_time{  # Take #Time from $_, return $x, $delta, $last_ctime
   if (/(^\d+):(\d+):(\d+) +\w+ +(\d+) +(\w+) +(\d+) +/) {
      #    HH    MM   SS    day    dd    mm     yyyy
#      $c_time = timegm($3,$2,$1, $4,$::month{$5}-1,$6-1900);
      $c_time = timelocal($3,$2,$1, $4,$::month{$5}-1,$6-1900);
#      print $_, "  ", $c_time, " ", $last_c_time, "\n";
      if ($last_c_time != 0) {
         my $interval = $c_time - $last_c_time;
         $delta = $interval - $c_interval;
         if (abs($delta) > 2) {
            print STDERR $_, "\n";
            print STDERR "   interval = $interval <<<<<<<<<\n";
            }
         }
      $last_c_time = $c_time;
#      $x = sprintf("%02d%02d.%02d%02d", $::month{$5},$4, $1,$2);
#      }               # mmdd.HHMM
      $x = sprintf("%04d%02d%02d.%02d%02d", $6,$::month{$5},$4, $1,$2);
      }               # YYYYmmdd:HHMM
   }


sub percentile{  # Compute $percent percentile of $dist,$nb,$tot
   my ($dist, $nb, $tot, $percent) = @_;  # Arguments

   if ($tot == 0) { return '?'; }

   my ($j, $t, $p, $lp, $c, $cj);
   $t = $p = $lp = 0;  $cj = -1;
   for ($j = 0; $j <= $nb; ++$j) {
      if (($c = $dist->{Dist}->[$j]) > 0) {
	 $t += $c;
         $p = $t * 100 / $tot;
         last if $p >= $percent;
         $lp = $p;  $cj = $j;
         }
      }
   if ($cj < 0) {  # No counts below $percent
#         ($t == $tot && $j == $cj+1)) {  # No counts above $percent
#printf "   cj=$cj, j=$j, lp=$lp, p=$p,  percent=$percent\n"; 
#      return -1;
      return $j;
      }
   return $j - ($j-$cj)*($p-$percent)/($p-$lp);
   }

sub put_stats{
   my($FILE, $var, $x, $first,$last, $percent) = @_;  # Arguments

   my ($ld, $div, $dist, $count);
   my (@min, @mean, @max, @p_lo, @p_med, @p_hi, @p_5, @p_95, 
      @lostpc, @totcnt);
   my (@low, @nb, @step);

   if ($var =~ /flowpdus|ime4/) {
      $div = 1;  # pkts (rather than kpkts)
                 # tat4 used for pkttrainlength
      }
   elsif ($var =~ /bitrate/) {
      $div =  1000000;  # Mbps (rather than kbps)
      }
   else {
      $div = 1000;  # kB/s, ms, kB ..
      }
#!print "first=$first, last=$last, div=$div\n";

   for ($count = 0, $ld = $first; $ld <= $last; ++$ld) {
      if (defined $distrib[$ld]) {
         my ($scale, $low, $high, $nb, $step);
         $dist = $distrib[$ld];
         if ($dist->{Trans} == 0) { next; }
         ++$count;
         $scale =  $dist->{Scale};
         $low = $dist->{Lower}*10**$scale;
         $high = $dist->{Upper}*10**$scale;
         $nb = $dist->{Buckets};
         if ($percent eq '/') {  # Ignore the 'overflow' counter
	    --$nb;
	    }
         if ($dist->{Trans} == 1) {
            $step = ($high - $low) / ($nb - 1);
            }
         else {
            $step = (log($high) - log($low)) / (($nb - 1)*$::ln10);
            }
         $low[$ld] = $low;  $nb[$ld] = $nb;  $step[$ld] = $step;
#!print "ld=$ld, low=$low, high=$high, nb=$nb,step=$step\n";
         }
      }
   
   if ($count == 0) {
      printf $FILE "#$x: no distribs in sample\n";
      return;
      }

   for (my $j = $first; $j <= $last; ++$j) {
      if (defined $distrib[$j]) {
 if ($nbr > 999999) {
print "distrib[$j] defined ...\n";
}
         $dist = $distrib[$j];
         if ($dist->{Trans} == 0) {
	    $p_lo[$j] = $mean[$j] = $p_hi[$j] = ' ?';
            next;
	    }

         my ($av, $tot, $mn, $mx, $p_med, $p_lo, $p_hi, $p_5, $p_95);
         for ($mn = 0; $mn <= $nb[$j]; ++$mn) {
            last if $dist->{Dist}->[$mn] != 0;
            }
         
         for (my $i = $mn, $mx = $tot = $av = 0; $i <= $nb[$j]; ++$i) {
            my $c = $dist->{Dist}->[$i];
            $tot += $c;
            $av += $c * $i;
            if ($c != 0) 
               { $mx = $i; }
            }
         $totcnt[$j] = sprintf " %d", $tot;
if ($totcnt[$j] < 0) {
     print "x=$x, j=$j, totcnt[j]=$totcnt[$j], tot=$tot\n";
   }
         if ($tot == 0) {  # Empty distribution 
	    $p_med[$j] = $p_lo[$j] =  $p_hi[$j] = 
               $mean[$j] = $min[$j] = $max[$j] = 
               $p_5[$j] = $p_95[$j] = ' ?';
            if ($lprq && defined($tolostx)) {
               $lostpc[$j] = $tolost[$j] == 0 ? " 0" : " 100";
               }
            next;
	    }
         elsif ($lprq && defined($tolostx)) {
	    my $lpc =  100*$tolost[$j]/($tolost[$j]+$tot);
	    $lostpc[$j] =  sprintf " %.2f", $lpc;
	    }

         if ($dist->{Trans} == 1) {
            $mn = $low[$j] + $mn*$step[$j];
            $av = $low[$j] + ($av/$tot)*$step[$j];
            $mx = $low[$j] + $mx*$step[$j];
            }
         else {
            $mn = $low[$j] * 10**($step[$j]*$mn);
            $av = $low[$j] * 10**($step[$j]*$av/$tot);
            $mx = $low[$j] * 10**($step[$j]*$mx);
            }
         $min[$j] = sprintf " %.2f", $mn/$div;
         $mean[$j] = sprintf " %.2f", $av/$div;
         $max[$j] = sprintf " %.2f", $mx/$div;

 if ($nbr > 999999) {
#!print "j=$j, mn=$mn, av=$av, mx=$mx, tot=$tot, nb=$nb[$j]\n";
print "nbr=$nbr, j=$j, mn=$mn, av=$av, mx=$mx, tot=$tot, nb=$nb[$j]\n";
}

         if ($tot == 0) {  # Not enough points to get median
	    $p_med[$j] = ' ?';
            next;
	    }
         else {
            if ($mdrq) {
               $p_med = percentile($dist, $nb[$j], $tot, 50);  # Median
               if ($p_med < 0) {
#  print "time=$x, j=$j, tot=$tot, median returned -1\n";
#   ++$nbr;
# if ($nbr > 5) { exit; }
                  $p_med[$j] = ' ?';
                  }
               else {
                  if ($dist->{Trans} == 1) {
                     $p_med = $low[$j] + $p_med*$step[$j];
                     }
                  else {
                     $p_med = $low[$j] * 10**($p_med*$step[$j]);
                     }
                  $p_med[$j] = sprintf " %.2f", $p_med/$div;
                  }
               }
            }
#!print "j=$j, median=$p_med[$j]\n";
         if ($tot < $min_pts) {  # Not enough points to get percentiles
	    $p_lo[$j] = $p_hi[$j] = ' ?';
            next;
	    }
      
         if ($lorq) {
            $p_lo = percentile($dist, $nb[$j], $tot, $lo_p);
            if ($p_lo < 0) { $p_lo[$j] = ' ?'; }
            else {
               if ($dist->{Trans} == 1) {
                  $p_lo = $low[$j] + $p_lo*$step[$j];
                  }
               else {
                  $p_lo = $low[$j] * 10**($p_lo*$step[$j]);
                  }
               $p_lo[$j] = sprintf " %.2f", $p_lo/$div;
               }
            }
         
         if ($hirq) {
            $p_hi = percentile($dist, $nb[$j], $tot, $hi_p);
            if ($p_hi < 0) { $p_hi[$j] = ' ?'; }
            else {
               if ($dist->{Trans} == 1) {
                  $p_hi = $low[$j] + $p_hi*$step[$j];
                  }
               else {
                  $p_hi = $low[$j] * 10**($p_hi*$step[$j]);
                  }
               $p_hi[$j] = sprintf " %.2f", $p_hi/$div;
               }
            }

         if ($tot < $min_5pts) {  # Not enough points to get 5%-iles
	    $p_5[$j] = $p_95[$j] = ' ?';
            next;
	    }
      
         if ($p5rq) {
            $p_5 = percentile($dist, $nb[$j], $tot, 5);
            if ($p_5 < 0) { $p_5[$j] = ' ?'; }
            else {
               if ($dist->{Trans} == 1) {
                  $p_5 = $low[$j] + $p_5*$step[$j];
                  }
               else {
                  $p_5 = $low[$j] * 10**($p_5*$step[$j]);
                  }
               $p_5[$j] = sprintf " %.2f", $p_5/$div;
               }
            }
         
         if ($p9rq) {
            $p_95 = percentile($dist, $nb[$j], $tot, 95);
            if ($p_95 < 0) { $p_95[$j] = ' ?'; }
            else {
               if ($dist->{Trans} == 1) {
                  $p_95 = $low[$j] + $p_95*$step[$j];
                  }
               else {
                  $p_95 = $low[$j] * 10**($p_95*$step[$j]);
                  }
               $p_95[$j] = sprintf " %.2f", $p_95/$div;
               }
            }
 if ($nbr > 999999) {
 print "nbr=$nbr, tot=$tot, min_5pts=$min_5pts, p_95=$p_95\n";
}
### if ($nbr >= 15) { exit; }
         }
      else {  # Missing disribution
         $min[$j] = $mean[$j] = $max[$j] = 
            $p_med[$j] = $p_lo[$j] = $p_hi[$j] = 
            $lostpc[$j] = $p_5[$j] = $p_95[$j] = $totcnt[$j] = ' ?';
         }
      }

   printf $FILE "$x ";

   if ($mdrq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $p_med[$j];
         }
      }
   if ($lprq && defined($tolostx)) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $lostpc[$j];
         }
      }
   if ($lorq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $p_lo[$j];
         }
      }
   if ($hirq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $p_hi[$j];
         }
      }
   if ($mnrq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $min[$j];
         }
      }
   if ($avrq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $mean[$j];
         }
      }
   if ($mxrq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $max[$j];
         }
      }
   if ($p5rq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $p_5[$j];
         }
      }
   if ($p9rq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $p_95[$j];
         }
      }
   if ($tcrq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         print $FILE $totcnt[$j];
         }
      }
   if ($tlrq) {
      printf $FILE " ";
      for (my $j = $first; $j <= $last; ++$j) {
         if (defined($tolost[$j])) { print $FILE " $tolost[$j]"; }
         else { print $FILE ' ?'; }
         }
      }
   print $FILE "\n";
   }

sub put_scalars{
   my($FILE, $x, $first,$last) = @_;  # Arguments

   print $FILE "$x ";
   for (my $j = $first; $j <= $last; ++$j) {
      if (defined $distrib[$j]) { print $FILE " $distrib[$j]"; }
      else { print $FILE ' ?'; }
      }
   print $FILE "\n";
   }
